@@ -94,6 +94,35 @@ void kill()
 void swap(ZZ& x)
 { _ntl_gswap(&rep, &x.rep); }
 
+bool pinned() const
+{
+   return _ntl_PINNED(rep);
+}
+
+
+#if (NTL_CXX_STANDARD >= 2011)
+
+ZZ(ZZ&& a) NTL_FAKE_NOEXCEPT
+{
+   *this = std::move(a);
+}
+
+ZZ& operator=(ZZ&& a) NTL_FAKE_NOEXCEPT
+{
+   if (pinned() || a.pinned()) {
+      _ntl_gcopy(a.rep, &rep);
+   }
+   else {
+      rep.move(a.rep);
+   }
+
+   return *this;
+}
+
+
+#endif
+
+
 void SetSize(long k)
 // pre-allocates space for k-digit numbers (base 2^NTL_ZZ_NBITS);  
 // does not change the value.
@@ -146,6 +175,9 @@ long validate() { return _ntl_gvalidate(rep); }
 };
 
 
+NTL_DECLARE_RELOCATABLE((ZZ*))
+
+
 class ZZWatcher {
 public:
    ZZ& watched;
@@ -714,6 +746,35 @@ struct sp_ZZ_reduce_struct {
 };
 
 
+// special-purpose routines for accumulating CRT-like summations
+// Not documented for now.
+
+
+// Allocates sz+2 limbs and zeros them all out.
+// x is not normalized.
+inline
+void QuickAccumBegin(ZZ& x, long sz)
+{
+   _ntl_quick_accum_begin(&x.rep, sz);
+}
+
+// x += y*b. 
+// Assumes y >= 0 and that 0 <= b < NTL_SP_BOUND.
+// Caller must assure that x does not exceed sz+2 limbs.
+// x remains unnormalized.
+inline
+void QuickAccumMulAdd(ZZ& x, const ZZ& y, long b)
+{
+   _ntl_quick_accum_muladd(x.rep, y.rep, b);
+}
+
+// renormalizes x.
+inline
+void QuickAccumEnd(ZZ& x)
+{
+   _ntl_quick_accum_end(x.rep);
+}
+
 
 /**********************************************************
 
@@ -948,7 +1009,9 @@ inline ZZ& operator^=(ZZ& x, long b)
 
 
 
-long NumBits(long a);
+inline
+long NumBits(long a)
+   { return _ntl_g2logs(a); }
 
 long bit(long a, long k);
 
@@ -1010,7 +1073,98 @@ void DeriveKey(unsigned char *key, long klen,
 
 #define NTL_PRG_KEYLEN (32)
 
+
+class RandomStream_impl;
+
+RandomStream_impl *
+RandomStream_impl_build(const unsigned char *key);
+
+RandomStream_impl *
+RandomStream_impl_build(const RandomStream_impl&);
+
+void
+RandomStream_impl_copy(RandomStream_impl&, const RandomStream_impl&);
+
+const unsigned char *
+RandomStream_impl_get_buf(const RandomStream_impl&);
+
+long
+RandomStream_impl_get_buf_len(const RandomStream_impl&);
+
+long
+RandomStream_impl_get_bytes(RandomStream_impl& impl, unsigned char *res, 
+   long n, long pos);
+
+void
+RandomStream_impl_delete(RandomStream_impl*);
+
+struct 
+RandomStream_impl_deleter {
+   static void deleter(RandomStream_impl* p) { RandomStream_impl_delete(p); }
+};
+
+
 class RandomStream {
+private:
+
+   long pos;
+   const unsigned char *buf;
+   long buf_len;
+
+   UniquePtr<RandomStream_impl,RandomStream_impl_deleter> impl;
+
+public:
+
+   explicit
+   RandomStream(const unsigned char *key) {
+      impl.reset(RandomStream_impl_build(key));
+      pos = buf_len = RandomStream_impl_get_buf_len(*impl);
+      buf = RandomStream_impl_get_buf(*impl);
+   }
+
+   RandomStream(const RandomStream& other) {
+      impl.reset(RandomStream_impl_build(*other.impl));
+      pos = other.pos;
+      buf_len = other.buf_len;
+      buf = RandomStream_impl_get_buf(*impl);
+   }
+
+   RandomStream& operator=(const RandomStream& other) {
+      RandomStream_impl_copy(*impl, *other.impl);
+      pos = other.pos;
+      buf_len = other.buf_len;
+      buf = RandomStream_impl_get_buf(*impl);
+      return *this;
+   }
+
+   void get(unsigned char *res, long n) 
+   {
+      // optimize short reads
+      if (n > 0 && n <= buf_len-pos) {
+#if 1
+         std::memcpy(&res[0], &buf[pos], n);
+         // NOTE: mempy undefined if res == null
+         // That's a reason we don't do this for n==0, which
+         // is anyway an unlikely case
+#else
+         for (long i = 0; i < n; i++) {
+            res[i] = buf[pos+i];
+         }
+#endif
+         pos += n;
+      }
+      else {
+         pos = RandomStream_impl_get_bytes(*impl, res, n, pos);
+      }
+   }
+
+};
+
+
+
+// ============ old stuff: for testing  ============
+
+class old_RandomStream {
 private:
    _ntl_uint32 state[16];
    unsigned char buf[64];
@@ -1020,7 +1174,7 @@ private:
 
 public:
    explicit
-   RandomStream(const unsigned char *key);
+   old_RandomStream(const unsigned char *key);
 
    // No default constructor 
    // default copy and assignment
